use std::str::FromStr;

use clap::Parser;
use cli::{Cli, Command, ProfileCommand};
use colored::Colorize;
use storage::Profile;
use tracing::Level;
use uonet::account::VulcanAccount;

mod cli;
mod storage;
mod sync;
mod term;

#[tokio::main(flavor = "current_thread")]
async fn main() {
	if let Err(e) = app().await {
		term::error(e.to_string().as_str())
	}
}

async fn app() -> anyhow::Result<()> {
	let cli = Cli::parse();
	tracing_subscriber::fmt::fmt()
		.with_max_level(Level::from_str(&cli.log)?)
		.init();

	if let Command::Profile { command } = cli.command {
		match command {
			ProfileCommand::List => {
				let (selected, other) = Profile::list()?;
				if other.is_empty() && selected.is_none() {
					term::frog("No profiles found");
				} else {
					term::header("Profiles:");
				}
				if let Some(selected) = selected {
					println!(" -> {} [selected]", selected.green());
				}
				for profile in other {
					println!(" -> {}", profile.bright_magenta());
				}
			}
			ProfileCommand::Select { name } => {
				Profile::select(&name)?;
				term::done(format!("Profile '{}' selected", name).as_str());
			}
			ProfileCommand::Register {
				name,
				noset,
				token,
				symbol,
				pin,
				device_name,
			} => {
				term::process("Registering Vulcan account");
				let account = VulcanAccount::register(&token, &symbol, &pin, device_name).await?;
				let creds = account.into_credentials();
				Profile::create(&name, creds)?;
				if !noset {
					Profile::select(&name)?;
				}
				term::done(format!("Profile '{}' registered", name).as_str());
			}
			ProfileCommand::Delete { name } => {
				Profile::delete(&name)?;
				term::done(format!("Profile '{}' deleted", name).as_str());
			}
		}
		return Ok(());
	}

	let mut profile = Profile::load_selected()?;

	if let Command::Sync = cli.command {
		sync::sync(&mut profile, true).await?;
		return Ok(());
	}
	sync::sync(&mut profile, false).await?;

	if let Command::Student { command } = cli.command {
		match command {
			cli::StudentCommand::List => {
				let current_student = profile.student;
				term::header("Students:");
				for student in profile.students {
					let pupil = student.pupil;
					if current_student == Some(pupil.id) {
						println!(
							" -> ({}) {} {} {} [selected]",
							pupil.id.to_string().as_str().bright_yellow(),
							pupil.first_name.bright_green(),
							pupil.surname.bright_green(),
							student.class_display.bright_cyan()
						);
					} else {
						println!(
							" -> ({}) {} {} {}",
							pupil.id.to_string().as_str().bright_yellow(),
							pupil.first_name.bright_green(),
							pupil.surname.bright_green(),
							student.class_display.bright_cyan()
						);
					}
				}
			}
			cli::StudentCommand::Select { id } => {
				let student = profile.students.iter().find(|v| v.pupil.id == id);
				if let Some(student) = student {
					term::done(
						format!(
							"Selected student {} {}",
							student.pupil.first_name.bright_yellow(),
							student.pupil.surname.bright_yellow()
						)
						.as_str(),
					);
					profile.student = Some(student.pupil.id);
					profile.commit()?;
				} else {
					term::error(format!("Student {} not found", id).as_str());
				}
			}
		}
		return Ok(());
	}

	//let student = profile.student.ok_or_else(|| anyhow!("No student selected."))?;

	Ok(())
}
