use std::env::current_dir;
use std::fs::{read_dir, remove_file, rename, File, OpenOptions};
use std::{fs::create_dir_all, path::PathBuf};

use anyhow::anyhow;
use serde::{Deserialize, Serialize};
use std::io::{self, ErrorKind, Read};
use uonet::account::Credentials;
use uonet::models::Student;

#[derive(Serialize, Deserialize, Debug)]
pub struct Profile {
	pub name: String,
	pub credentials: Credentials,
	pub students: Vec<Student>,
	pub student: Option<i32>,
	pub last_sync: u64,
}

fn config_dir() -> io::Result<PathBuf> {
	let path = dirs::config_dir();
	if let Some(path) = path {
		let path = path.join("uonet-cli");
		create_dir_all(&path)?;
		Ok(path)
	} else {
		Ok(current_dir()?.join("uonet-cli"))
	}
}

impl Profile {
	/// Updates the profile on disk
	pub fn commit(&self) -> anyhow::Result<()> {
		let path = config_dir()?.join(format!("_{}", self.name));
		let file = File::create(path)?;
		serde_json::to_writer(file, self)?;
		Ok(())
	}

	/// Creates a profile from saved Vulcan credentials
	pub fn create(name: &str, credentials: Credentials) -> anyhow::Result<Self> {
		if name.starts_with("_") {
			return Err(anyhow!("Profile name cannot start with an underscore"));
		}
		let dir = config_dir()?;
		if dir.join(format!("_{}", name)).exists() {
			return Err(anyhow!("Profile '{}' already exists and is selected", name));
		}
		let file = match OpenOptions::new()
			.write(true)
			.create_new(true)
			.open(dir.join(name))
		{
			Ok(o) => o,
			Err(e) => {
				return if e.kind() == ErrorKind::AlreadyExists {
					return Err(anyhow!("Profile '{}' already exists", name));
				} else {
					Err(anyhow!(e))
				}
			}
		};
		let profile = Profile {
			credentials,
			name: name.to_string(),
			student: None,
			students: vec![],
			last_sync: 0,
		};
		serde_json::to_writer(file, &profile)?;
		Ok(profile)
	}

	/// Deletes a profile by name
	pub fn delete(name: &str) -> anyhow::Result<()> {
		let dir = config_dir()?;
		let result =
			remove_file(dir.join(format!("_{}", name))).or_else(|_| remove_file(dir.join(name)));
		if let Err(e) = result {
			if e.kind() == ErrorKind::NotFound {
				Err(anyhow!("Profile '{}' does not exist", name))
			} else {
				Err(anyhow!(e))
			}
		} else {
			Ok(())
		}
	}

	/// Lists all available profiles. Returns the selected profile and a list of unselected profiles.
	pub fn list() -> io::Result<(Option<String>, Vec<String>)> {
		let files = read_dir(config_dir()?)?;
		let mut selected = None;
		let mut other = vec![];
		for file in files {
			let filename = file?.file_name();
			let filename = filename.to_string_lossy();
			if filename.starts_with("_") {
				selected = Some((&filename[1..]).to_string());
			} else {
				other.push(filename.into_owned())
			}
		}
		Ok((selected, other))
	}

	/// Selects a given profile by name.
	pub fn select(name: &str) -> anyhow::Result<()> {
		let dir = config_dir()?;
		let selecting_oldname = dir.join(name);
		if !selecting_oldname.is_file() {
			return Err(anyhow!("Profile {} does not exist.", name));
		}
		let selecting_newname = dir.join(format!("_{}", name));
		let files = read_dir(&dir)?;
		for file in files {
			let filename = file?.file_name();
			let filename = filename.to_string_lossy();
			if filename.starts_with("_") {
				let selected_oldname = dir.join(filename.as_ref());
				let selected_newname = dir.join(&filename[1..]);
				rename(selected_oldname, selected_newname)?;
			}
		}
		rename(selecting_oldname, selecting_newname)?;
		Ok(())
	}

	/// Loads currently selected profile.
	pub fn load_selected() -> anyhow::Result<Profile> {
		let files = read_dir(config_dir()?)?;
		for file in files {
			let file = file?;
			let name = file.file_name();
			if name.to_string_lossy().starts_with("_") {
				let mut v = vec![];
				File::open(file.path())?.read_to_end(&mut v)?;
				let storage: Profile = serde_json::from_slice(&v)?;
				return Ok(storage);
			}
		}
		Err(anyhow!("No profile selected. Create a profile first."))
	}
}
