use crate::error::Result;
use crate::models::{Period, RegisterResponse, Student};
use crate::request::{read_uonet_response, UonetRequest};
use crate::UonetError;
use openssl::asn1::{Asn1Integer, Asn1Time};
use openssl::base64;
use openssl::bn::BigNum;
use openssl::hash::MessageDigest;
use openssl::nid::Nid;
use openssl::pkey::{PKey, Private};
use openssl::rsa::Rsa;
use openssl::x509::{X509NameBuilder, X509};
use reqwest::Client;
use serde::{Deserialize, Serialize};
use serde_json::json;
use uuid::Uuid;

/// API client containing the credentials
pub struct VulcanAccount {
	pub(crate) credentials: Credentials,
	pub(crate) http: Client,
	pub(crate) pkey: PKey<Private>,
	pub(crate) student: Option<Student>,
	pub(crate) period: Option<Period>,
}

impl VulcanAccount {
	/// Registers a new account using the codes provided on the "Dostęp mobilny" page in UONET+.
	pub async fn register(
		token: &str,
		symbol: &str,
		pin: &str,
		device_name: String,
	) -> Result<Self> {
		for b in device_name.as_bytes() {
			if !(*b >= 32 && *b != 127 || *b == b'\t') {
				return Err(UonetError::InvalidDeviceName(device_name));
			}
		}

		let token = token.to_uppercase();
		let symbol = symbol.to_ascii_lowercase();

		let http = Client::new();
		let base_url = get_base_path(&http, &token).await?;
		let fbresp = get_firebase_token(&http).await?;
		let firebase_token = fbresp
			.strip_prefix("token=")
			.ok_or_else(|| UonetError::InvalidFirebaseToken((&fbresp).clone()))?
			.to_string();

		let pkey = PKey::from_rsa(Rsa::generate(2048)?)?;

		let serial = Asn1Integer::from_bn(BigNum::from_u32(1)?.as_ref())?;
		let notbefore = Asn1Time::days_from_now(0)?;
		let notafter = Asn1Time::days_from_now(365 * 8)?;

		let mut sname = X509NameBuilder::new()?;
		sname.append_entry_by_nid(Nid::COMMONNAME, "APP_CERTIFICATE CA Certificate")?;
		let sname = sname.build();

		let mut x509 = X509::builder()?;
		x509.set_version(2)?;
		x509.set_serial_number(&serial)?;
		x509.set_issuer_name(&sname)?;
		x509.set_pubkey(&pkey)?;
		x509.set_not_before(&notbefore)?;
		x509.set_not_after(&notafter)?;
		x509.sign(&pkey, MessageDigest::sha256())?;
		let x509 = x509.build();

		let cert = unsafe { String::from_utf8_unchecked(x509.to_pem()?).replace('\n', "") };
		let cert = cert.split("-----").nth(2).unwrap().to_string();
		let key = unsafe {
			String::from_utf8_unchecked(pkey.private_key_to_pem_pkcs8()?).replace('\n', "")
		};
		let key = key.split("-----").nth(2).unwrap().to_string();
		let fingerprint = hex::encode(x509.digest(MessageDigest::sha1())?);

		let uuid = Uuid::new_v5(&Uuid::NAMESPACE_X500, fingerprint.as_bytes())
			.as_hyphenated()
			.to_string();

		let body = json!({
			"OS": "Android",
			"DeviceModel": device_name,
			"Certificate": cert,
			"CertificateType": "X509",
			"CertificateThumbprint": fingerprint,
			"PIN": pin,
			"SecurityToken": token,
			"SelfIdentifier": uuid,
		});

		let full_url = format!("{}/{}/api/mobile/register/new", base_url, symbol);

		let credentials = Credentials {
			cert,
			key,
			fingerprint,
			firebase_token,
			device_name,
			rest_url: base_url,
			login_id: 0,
		};

		#[cfg(feature = "tracing")]
		tracing::debug!("generated {:?}", credentials);

		let mut acc = VulcanAccount {
			credentials,
			http,
			pkey,
			student: None,
			period: None,
		};

		let req = acc
			.http
			.post(&full_url)
			.envelope(&full_url, &acc, Some(body))?
			.build()?;
		let resp = acc.http.execute(req).await?.text().await?;
		let resp: RegisterResponse = read_uonet_response(&resp)?;

		acc.credentials.rest_url = resp.rest_url;
		acc.credentials.login_id = resp.login_id;

		#[cfg(feature = "tracing")]
		tracing::info!("Logged in as '{}'", resp.user_name);

		Ok(acc)
	}

	/// Creates a VulcanAccount from saved credentials.
	pub fn from_credentials(credentials: Credentials) -> Result<Self> {
		let der = base64::decode_block(&credentials.key)?;
		Ok(Self {
			credentials,
			http: Client::new(),
			pkey: PKey::private_key_from_pkcs8(&der)?,
			student: None,
			period: None,
		})
	}

	/// Gets saved credentials from the account.
	pub fn credentials(&self) -> &Credentials {
		&self.credentials
	}

	/// Converts the account to owned credentials.
	pub fn into_credentials(self) -> Credentials {
		self.credentials
	}

	/// Gets the device name.
	pub fn device_name(&self) -> &str {
		&self.credentials.device_name
	}

	/// Gets a list of students from the API. You can then set the `student` field to one of those.
	pub async fn list_students(&self) -> Result<Vec<Student>> {
		let url = format!("{}api/mobile/register/hebe", self.credentials.rest_url);
		let req = self.http.get(&url).envelope(&url, &self, None)?.build()?;
		let resp = self.http.execute(req).await?.text().await?;
		Ok(read_uonet_response(&resp)?)
	}

	/// Sets the student object used in queries.
	///
	/// You can technically load a deserialized Student here, although it's not recommended,
	/// as I have absolutely no idea what happens when you use an invalid Student object in an API call.
	pub fn set_student(&mut self, student: Student) -> Result<()> {
		let current_period = student
			.periods
			.iter()
			.find(|v| v.current)
			.ok_or(UonetError::InvalidStudent)?;
		self.period = Some(current_period.clone());
		self.student = Some(student);
		Ok(())
	}

	/// Gets the currently set student object.
	pub fn get_student(&self) -> Option<&Student> {
		self.student.as_ref()
	}

	/// Sets the period used in queries.
	pub fn set_period(&mut self, period: Period) {
		self.period = Some(period);
	}

	/// Gets the currently set period.
	pub fn get_period(&self) -> Option<&Period> {
		self.period.as_ref()
	}
}

/// Credentials for [`VulcanAccount`].
///
/// You can get them from an account using either [`credentials`](VulcanAccount::credentials) or [`into_credentials`](VulcanAccount::into_credentials).
///
/// This type implements Serialize and Deserialize, so you can save the credentials to a file, for example.
#[derive(Serialize, Deserialize, Debug, Clone)]
pub struct Credentials {
	pub(crate) cert: String,
	pub(crate) key: String,
	pub(crate) fingerprint: String,
	pub(crate) firebase_token: String,
	pub(crate) device_name: String,
	pub(crate) rest_url: String,
	pub(crate) login_id: i32,
}

async fn get_firebase_token(http: &Client) -> std::result::Result<String, reqwest::Error> {
	const FIREBASE_BODY: &[(&str, &str)] = &[
		("sender", "987828170337"),
		("X-scope", "*"),
		("X-gmp_app_id", "1:987828170337:android:ac97431a0a4578c3"),
		("app", "pl.edu.vulcan.hebe"),
		("device", "4609707972546570896"),
	];

	let req = http
		.post("https://android.clients.google.com/c2dm/register3")
		.header(
			"Authorization",
			"AidLogin 4609707972546570896:3626695765779152704",
		)
		.header("User-Agent", "Android-GCM/1.5")
		.header("app", "pl.edu.vulcan.hebe")
		.form(FIREBASE_BODY)
		.build()?;

	Ok(http.execute(req).await?.text().await?)
}

async fn get_base_path(http: &Client, token: &str) -> Result<String> {
	let req = http
		.get("http://komponenty.vulcan.net.pl/UonetPlusMobile/RoutingRules.txt")
		.build()?;
	let resp = http.execute(req).await?;
	let content_type = resp
		.headers()
		.get("Content-Type")
		.ok_or(UonetError::IncompleteResponse(None))?
		.to_str()
		.map_err(|_| UonetError::IncompleteResponse(None))?;
	if content_type == "text/plain" {
		let text = resp.text().await?;
		let shortok = &token[..3];
		for line in text.lines() {
			let components = line
				.split_once(',')
				.ok_or_else(|| UonetError::IncompleteResponse(Some(text.clone())))?;
			if components.0 == shortok {
				return Ok(components.1.to_string());
			}
		}
	}
	Err(UonetError::InvalidToken)
}
