use crate::account::VulcanAccount;
use crate::UonetError;
use crate::models::UonetResponse;
use crate::UonetError::ErrorResponse;
use chrono::{Local, Utc};
use openssl::base64;
use openssl::error::ErrorStack;
use openssl::hash::MessageDigest;
use openssl::sha::sha256;
use openssl::sign::Signer;
use percent_encoding::{utf8_percent_encode, NON_ALPHANUMERIC};
use reqwest::header::{HeaderMap, HeaderValue};
use reqwest::RequestBuilder;
use serde::Deserialize;
use serde_json::{json, Value};
use uuid::Uuid;

pub(crate) trait UonetRequest {
	fn envelope(
		self,
		url: &str,
		account: &VulcanAccount,
		body: Option<Value>,
	) -> Result<Self, UonetError>
	where
		Self: Sized;
}

impl UonetRequest for RequestBuilder {
	// Adds headers, envelopes and signs a UONET+ API request
	fn envelope(
		self,
		url: &str,
		account: &VulcanAccount,
		body: Option<Value>,
	) -> Result<Self, UonetError> {
		let canonical_url = get_canonical_url(url).ok_or_else(|| UonetError::InvalidUrl(url.to_string()))?;
		let vdate = Utc::now().format("%a, %d %b %Y %H:%M:%S GMT").to_string();
		let mut headers = HeaderMap::with_capacity(10);

		headers.insert(
			"User-Agent",
			HeaderValue::from_static("Dart/2.10 (dart:io)"),
		);
		headers.insert("vOS", HeaderValue::from_static("Android"));
		headers.insert("vAPI", HeaderValue::from_static("1"));

		headers.insert(
			"vDeviceModel",
			HeaderValue::from_str(&account.credentials.device_name).unwrap(),
		);
		headers.insert("vDate", HeaderValue::from_str(&vdate).unwrap());
		headers.insert(
			"vCanonicalUrl",
			HeaderValue::from_str(&canonical_url).unwrap(),
		);

		let mut req = self;
		if let Some(body) = body {
			let uuid = Uuid::new_v4().as_hyphenated().to_string();
			let local = Local::now();
			let local_millis = local.timestamp_millis();
			let local_iso = local.format("%Y-%m-%d %H:%M:%S").to_string();
			let payload = json!({
				"AppName": "DzienniczekPlus 2.0",
				"AppVersion": "1.4.2",
				"CertificateId": account.credentials.fingerprint,
				"Envelope": body,
				"FirebaseToken": account.credentials.firebase_token,
				"API": 1,
				"RequestId": uuid,
				"Timestamp": local_millis,
				"TimestampFormatted": local_iso
			});
			let payload_str = payload.to_string();
			let digest = base64::encode_block(&sha256(payload_str.as_bytes()));
			let digest_header = format!("SHA-256={}", digest);
			let signature = get_signature_header(account, &canonical_url, &vdate, Some(&digest))?;
			headers.insert("Signature", HeaderValue::from_str(&signature).unwrap());
			headers.insert("Digest", HeaderValue::from_str(&digest_header).unwrap());
			headers.insert("Content-Type", HeaderValue::from_static("application/json"));
			req = req.body(payload_str);
		} else {
			let signature = get_signature_header(account, &canonical_url, &vdate, None)?;
			headers.insert("Signature", HeaderValue::from_str(&signature).unwrap());
		}

		Ok(req.headers(headers))
	}
}

fn get_canonical_url(full_url: &str) -> Option<String> {
	let mut url = utf8_percent_encode(&full_url[full_url.find("api/mobile/")?..], NON_ALPHANUMERIC)
		.to_string();
	url.make_ascii_lowercase();
	Some(url)
}

fn get_signature_header(
	account: &VulcanAccount,
	canonical_url: &str,
	vdate: &str,
	digest: Option<&str>,
) -> Result<String, ErrorStack> {
	let (sign_headers, sign_values) = if let Some(d) = digest {
		(
			"vCanonicalUrl Digest vDate",
			format!("{}{}{}", canonical_url, d, vdate),
		)
	} else {
		("vCanonicalUrl vDate", format!("{}{}", canonical_url, vdate))
	};
	let mut signer = Signer::new(MessageDigest::sha256(), &account.pkey)?;
	let signature = base64::encode_block(
		signer
			.sign_oneshot_to_vec(sign_values.as_bytes())?
			.as_slice(),
	);
	Ok(format!("keyId=\"{}\",headers=\"{}\",algorithm=\"sha256withrsa\",signature=Base64(SHA256withRSA({}))", account.credentials.fingerprint, sign_headers, signature))
}

// un-envelope the api response
pub fn read_uonet_response<'a, T: Deserialize<'a>>(json: &'a str) -> Result<T, UonetError> {
	#[cfg(feature = "tracing")]
	tracing::debug!("API response: {}", json);
	let resp: UonetResponse<T> = serde_json::from_str(json)?;
	if resp.status.code != 0 {
		Err(ErrorResponse(resp.status))
	} else {
		Ok(resp
			.envelope
			.ok_or_else(|| UonetError::IncompleteResponse(Some(json.to_string())))?)
	}
}
