use chrono::{NaiveDate, NaiveDateTime};
use serde::de::DeserializeOwned;

use crate::{
	account::VulcanAccount,
	request::{read_uonet_response, UonetRequest},
	Result, UonetError,
};

pub enum FilterType {
	ByPupil,
	ByLoginId,
}

impl VulcanAccount {
	pub(crate) async fn get_list<T: DeserializeOwned>(
		&self,
		endpoint: &str,
		ft: FilterType,
		from: Option<NaiveDate>,
		to: Option<NaiveDate>,
		last_sync: Option<NaiveDateTime>,
	) -> Result<Vec<T>> {
		let student = self.student.as_ref().ok_or(UonetError::StudentMissing)?;
		let period = self.period.as_ref().unwrap();
		let mut query = Vec::with_capacity(8);
		let url = match ft {
			FilterType::ByPupil => {
				query.push(("unitId", student.unit.id.to_string()));
				query.push(("pupilId", student.pupil.id.to_string()));
				query.push(("periodId", period.id.to_string()));
				format!(
					"{}{}/api/mobile/{}/byPupil",
					self.credentials.rest_url, student.unit.symbol, endpoint
				)
			}
			FilterType::ByLoginId => {
				query.push(("loginId", self.credentials.login_id.to_string()));
				format!(
					"{}{}/api/mobile/{}",
					self.credentials.rest_url, student.unit.symbol, endpoint
				)
			}
		};
		if let Some(from) = from {
			query.push(("dateFrom", from.format("%Y-%m-%d").to_string()));
		}
		if let Some(to) = to {
			query.push(("dateTo", to.format("%Y-%m-%d").to_string()));
		}
		let last_sync = last_sync
			.map(|v| v.format("%Y-%m-%d %H:%m:%S").to_string())
			.unwrap_or_else(|| "1970-01-01 00:00:00".into());
		query.push(("pageSize", "500".to_string()));
		query.push(("lastId", "-2147483648".to_string()));
		query.push(("lastSyncDate", last_sync));

		let req = self
			.http
			.get(&url)
			.query(&query)
			.envelope(&url, self, None)?
			.build()?;
		let resp = self.http.execute(req).await?.text().await?;
		read_uonet_response(&resp)
	}
}
